import os
import time
import hashlib
import functools
import pprint
import json


class Receipt:
    '''Make a receipt for a function call. This allows you skip over a function if 
    it was successfully ran. This is useful if a long script fails in the middle and 
    you want to re-run it, but you don't need to re-run the first part. 

    This will cache the function execution history using the string representation 
    of the function's arguments. This works well as 99% of things have some sort of 
    string representation, however, if you have a string representation that doesn't
    stay consistent, like an object that prints out its ID, then the receipt won't work.

    To resolve this, you can either:

     * subclass this method and override ``Receipt.hash_args(*a, **kw)`` to return something invariant
       for your
     * Provide your own ``receipt_id`` to the function call.
     * submit a PR for a better hash function!
    
    .. code-block:: python

        receipt = Receipt(func.__name__, *a, **kw)

        if not receipt.exists():
            try:
                do_something()
                receipt.make()
                # yay we can skip it next time!
            except Exception:
                """Oh well. It failed, but we'll just try again next time."""

        else:
            """Oh good it ran completely last time so we can skip it and move to the next one."""
    '''
    ROOT_DIR = './receipts'
    TEST = False
    def __init__(self, name='', *a, receipt_id=None, __dir__=None, **kw):
        if callable(name):
            name = getattr(name, '__qualname__') or getattr(name, '__name__')
        assert name or a or kw, 'you must pass some identifiable information to be used for a hash.'
        self.name = name
        self.id = '{}{}'.format(name or '', receipt_id or self.hash_args(*a, **kw))
        self.ROOT_DIR = __dir__ or self.ROOT_DIR
        self.fname = os.path.join(self.ROOT_DIR, self.id)

    def __str__(self):
        return '<Receipt exists={} file={}>'.format(
            self.exists, self.fname)

    def hash_args(self, *a, **kw):
        '''Take the function arguments and return a hash string for them.'''
        return hashlib.md5((
            str(a) + str(sorted(kw.items()))
        ).encode()).hexdigest()

    @property
    def exists(self):
        return os.path.isfile(self.fname)

    def make(self, **meta):
        os.makedirs(self.ROOT_DIR, exist_ok=True)
        with open(self.fname, 'w') as f:
            os.utime(self.fname)
            try:
                json.dump(meta, f)
            except Exception as e:
                json.dump({
                    'error': type(e).__name__,
                    'description': str(e),
                    'data': str(meta)
                }, f)

    def clear(self):
        if self.exists:
            os.remove(self.fname)

    @property
    def meta(self):
        if os.path.isfile(self.fname):
            with open(self.fname, 'r') as f:
                try:
                    s = f.read()
                    return json.loads(s) if s else {}
                except json.decoder.JSONDecodeError as e:
                    print('error:', e, s)
                    return s



def _fallbacks(*xs):
    for x in xs:
        if x is not None:
            return x


def use_receipt(func, receipt_dir=None, test=None):
    '''Use a receipt for a function call, which lets us skip a result if the function completed successfully
    the last run. This is just a wrapper around ``Receipt`` that handles the receipt checking/making logic for you.

    .. code-block:: python

        # do step 1. If it already ran successfully it will skip and do nothing.
        use_receipt(my_step1_function)(**step1_kwargs)

        # do step 2. here we're passing a custom receipt ID
        custom_receipt_id = ...
        use_receipt(my_step2_function)(**step2_kwargs, receipt_id=custom_receipt_id)

        # do step 3
        use_receipt(my_step3_function)(**step3_kwargs)
    '''
    @functools.wraps(func)
    def inner(*a, overwrite_=False, test=None, receipt_dir_=None, **kw):
        test = _fallbacks(test, inner.TEST, Receipt.TEST)
        receipt_dir_ = _fallbacks(receipt_dir_, inner.ROOT_DIR, Receipt.ROOT_DIR)
        r = Receipt(func, *a, __dir__=receipt_dir_, **kw)
        name = r.name
        if test:
            print('''
------------------------
-- Test Run --

Function: {}
Receipt: {}
*args:
{}
**kwargs:
{}
------------------------
            '''.format(name, r, pprint.pformat(a), pprint.pformat(kw)))
            return
        if overwrite_ or not r.exists:
            start_time = time.time()
            try:
                result = func(*a, **kw)
            except BaseException as e:
                r.clear()
                print('''
------------------------
-- Error during receipted function {}. --
Receipt: {}
Error: ({}) {}

No receipt is written.
------------------------
                '''.format(name, r, type(e).__name__, e))
                raise
            r.make(duration_secs=time.time() - start_time, time=time.time())

            print('''
------------------------
-- Receipt written for {} --
Receipt: {}

Took: {} seconds.
------------------------
            '''.format(name, r, (r.meta or {}).get('duration_secs')))
            return result
        else:
            print('''
------------------------
-- Receipt exists for {} --
Receipt: {}

Skipping.
------------------------
            '''.format(name, r, (r.meta or {}).get('duration_secs')))
    inner.TEST, inner.ROOT_DIR = test, receipt_dir
    return inner


# so that setting attributes will set receipt instead
class _DeprecatedSetAttr:
    func = None
    def __init__(self, func):
        self.func = func
        functools.update_wrapper(self, func)

    def __call__(self, *a, **kw):
        return self.func(*a, **kw)

    def __setattr__(self, k, v):
        print(k, v, k not in self.__dict__)
        if k not in self.__dict__:
            setattr(Receipt, k, v)
            return
        super().__setattr__(k, v)



import os
import sys
import json
import shlex
import types
import itertools
import collections



def summary(run_script, job_paths):
    '''Print out a nice summary of the job sbatch files and the run script.'''
    print('Generated', len(job_paths), 'job scripts:')
    for p in job_paths:
        print('\t', p)
    print()

    print('To submit all jobs, run:')
    print('.', run_script)
    print()


'''

Misc / template utils

'''

def command_to_name(command):
    '''Convert a bash command into a usable job name.
    
    NOTE: This was originally to turn 'python my/script.py' into 'my.script'
          but obviously this is just a special case and I never really found
          a nice, general way to do this.
    '''
    fbase = os.path.splitext(shlex.split(command)[1])[0]
    return fbase.replace('/', '.').lstrip('.').replace(' ', '-')


def as_chunks(lst, n=1):
    return [lst[i:i + n] for i in range(0, len(lst), n)]



def make_executable(file_path):
    '''Grant permission to execute a file.'''
    # https://stackoverflow.com/a/30463972
    mode = os.stat(file_path).st_mode
    mode |= (mode & 0o444) >> 2
    os.chmod(file_path, mode)


def maybe_backup(path, prefix='~'):
    '''Backup a path if it exists.'''
    # create backup
    if path.exists():
        bkp_previous_path = path.prefix(prefix).next_unique(1)
        os.rename(path, bkp_previous_path)
        print('moved existing', path, 'to', bkp_previous_path)



def prettyjson(value):
    '''Pretty-print data using json.'''
    return json.dumps(value, sort_keys=True, indent=4) if value else ''


def prefixlines(text, prefix='# '):
    '''Prefix each line. Can be used to comment a block of text.'''
    return ''.join(prefix + l for l in str(text).splitlines(keepends=True))


def dict_merge(*ds, depth=-1, **kw):
    '''Recursive dict merge.

    Arguments:
        *ds (dicts): dicts to be merged.
        depth (int): the max depth to be merged.
        **kw: extra keys merged on top of the final dict.

    Returns:
        merged (dict).
    '''
    def merge(dicta, dictb, depth=-1):
        '''Recursive dict merge.

        Arguments:
            dicta (dict): dict to be merged into.
            dictb (dict): merged into dicta.
        '''
        for k in dictb:
            if (depth != 0 and k in dicta
                    and isinstance(dicta[k], dict)
                    and isinstance(dictb[k], collections.Mapping)):
                dict_merge(dicta[k], dictb[k], depth=depth - 1)
            else:
                dicta[k] = dictb[k]

    mdict = {}
    for d in ds + (kw,):
        merge(mdict, d, depth=depth)
    return mdict


def flatten(args, cls=(list, tuple, set, types.GeneratorType)):
    '''Flatten iterables into a flat generator.'''
    if isinstance(args, cls):
        yield from (a for arg in args for a in flatten(arg))
    elif args is not None:
        yield args


def clsattrdiff(cls, base_cls=None, funcs=False):
    '''Get the non-function attributes that exist on ``cls``, but not on ``base_cls``.'''
    attrs = dict()
    base_dict = base_cls.__dict__ if base_cls is not None else {}
    i = cls.__mro__.index(base_cls) if base_cls is not None else None
    for cls_i in cls.__mro__[:i][::-1]:
        for k, v in cls_i.__dict__.items():
            if (not k.startswith('_') and 
                    k not in base_dict and 
                    (funcs or not callable(v))):
                attrs[k] = v
    return attrs


import re
import unicodedata

_filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_.-]")
_windows_device_files = ("CON", "AUX", "COM1", "COM2", "COM3", "COM4", "LPT1", "LPT2", "LPT3", "PRN", "NUL")

def secure_filename(filename):
    r"""Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.
    On windows systems the function also makes sure that the file is not
    named after one of the special device files.
    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename('i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'
    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you abort or
    generate a random filename if the function returned an empty one.
    """
    filename = unicodedata.normalize("NFKD", filename).encode("ascii", "ignore").decode("ascii")

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, " ")
    filename = str(_filename_ascii_strip_re.sub(
        "", "_".join(filename.split())
    )).strip("._")

    # on nt a couple of special files are present in each folder.  Don't use those
    if (os.name == "nt" and filename and filename.split(".")[0].upper() in _windows_device_files):
        filename = f"_{filename}"
    return filename



'''

Argument

'''


def all_subclasses(cls):
    '''Return all recursive subclasses of a class.'''
    return set(cls.__subclasses__()).union(
        s for c in cls.__subclasses__() for s in all_subclasses(c))


def subclass_lookup(cls, prefix='', suffix='', name_override=None):
    '''Get a dictionary of name to subclass (removing any common prefix/suffix).'''
    return {
        (getattr(c, name_override, None) if name_override else None) 
            or stripstr(c.__name__.lower(), prefix, suffix): c
        for c in all_subclasses(cls)
        if not c.__name__.startswith('_')}


def shlex_repr(v):
    '''Prepare a variable for bash. This will serialize the object using json 
    and then quote the variable if it contains any spaces/bash-specific characters.
    '''
    v = json_safe_numpy(v)
    # v = repr(v)
    v = json.dumps(v)  # TODO: is there a better option?
    # if v is a quoted string, remove the quotes
    if len(v) >= 2 and v[0] == v[-1] and v[0] in '"\'':
        v = v[1:-1]
    return shlex.quote(v) # only quote if necessary (has spaces or bash chars)

def json_safe_numpy(obj):
    if 'numpy' in sys.modules:  # don't need to check if it's not imported
        import numpy as np
        if isinstance(obj, np.generic):
            return obj.item()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
    return obj


def escape(txt, chars='"\''):
    '''Escape certain characters from a string. By default, it's quotes.'''
    for c in chars:
        txt = txt.replace(c, '\\' + c)
    return txt


def stripstr(text, prefix='', suffix=''):
    '''Strip a prefix and/or suffix from a string.'''
    text = text[len(prefix):] if prefix and text.startswith(prefix) else text
    text = text[:-len(suffix)] if suffix and text.endswith(suffix) else text
    return text


def get_available_tensorflow_cuda_versions():
    '''Load the tensorflow-cuda lookup table.'''
    # https://www.tensorflow.org/install/source#gpu
    import csv
    with open(os.path.join(os.path.dirname(__file__), 'cuda_versions.csv'), 'r') as f:
        rows = list(csv.DictReader(f, delimiter='\t'))
    versions = ({
        'cuda': d['CUDA'], 'cudnn': d['cuDNN'], 
        'python': d["Python version"].split(', '),
        'version': d['Version'].split('-')[-1]
    } for d in rows)
    return {d['version']: d for d in versions}


def find_tensorflow_cuda_version(version, latest=False):
    '''Given a tensorflow version, lookup the correct'''
    import fnmatch
    versions = get_available_tensorflow_cuda_versions()
    version = str(version).split('.')[:2]
    version = '.'.join(version + ['*']*(3 - len(version)))
    matched = sorted(v for v in versions if fnmatch.fnmatch(v, version))
    if not matched:
        raise ValueError('No value matching {!r}'.format(version))
    return versions[matched[-1 if latest else 0]]

if __name__ == '__main__':
    import fire
    fire.Fire()


import os
import subprocess
import slurmjobs

OVERLAY_DIR = "/scratch/work/public/overlay-fs-ext3/"
SIF_DIR = "/scratch/work/public/singularity"
 
def singuconda(overlay, sif, package=None, requirements=None, install=None, conda_env=None, python_version=None,
               sif_dir=SIF_DIR, overlay_dir=OVERLAY_DIR, overlay_src=None):
    '''Installing conda in a singularity overlay.
    
    Arguments:
        overlay (str): The path to the local overlay file.
        sif (str): The name of the sif file (either relative to ``sif_dir`` or an absolute path)

        package (str): a local python package: a repo directory containing a ``setup.py`` file where your project source lives.
            i.e. ``pip install -e ./your/package`` - you provide: ``./your/package``
        requirements (str): a pip requirements file: 
            i.e. ``pip install -r ./your/requirements.txt`` - you provide: ``./your/requirements.txt``
        install (str, list): any conda packages that you want to install: conda packages
            i.e. ``conda install numpy librosa`` - you provide: ``numpy librosa``

        conda_env (str): The name of a conda environment. If neither this, nor python version is provided, 
            the base environment will be used. If this is not defined and a python version is specified, 
            the conda environment will be ``f'py{python_version}'``
        python_version (str): The python version to use.

        sif_dir (str): Where the sif files are stored. This is ignored if an absolute path is provided for ``sif``.
            This just makes it more convenient to specify a sif file, under the assumption that you're using 
            NYU's provided sifs.
        overlay_dir (str): Where the zipped overlays are stored.
        overlay_src (str): 

    
    '''
    overlay_src = os.path.join(overlay_dir or '', overlay_src or f'{overlay}.gz')
    subprocess.run(f'''
OVERLAY="{overlay}"
OVERLAY_SRC="{overlay_src}"
''' + '''
# get overlay
if [ ! -f "${OVERLAY}" ]; then
    echo pulling overlay file...
    [ -f "${OVERLAY}.gz" ] || cp -rp "$OVERLAY_SRC" "${OVERLAY}.gz"
    echo unzipping overlay file...
    gunzip "${OVERLAY}.gz"
fi
    ''')

    header('setting up conda')

    # overlay = os.path.join(overlay_dir or '', overlay)
    sif = os.path.join(sif_dir or '', sif)

    if python_version and not conda_env:
        conda_env = f'py{python_version}'

    sing(overlay, sif, f"""
CONDA_ENV="{conda_env or ''}"
CONDA_PYTHON="{f'python={python_version}' if python_version else ''}"
""" + """
if [ ! -e /ext3/miniconda3 ]; then
    echo installing miniconda inside container...
    [[ ! -f Miniconda3-latest-Linux-x86_64.sh ]] && wget https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86_64.sh
    bash Miniconda3-latest-Linux-x86_64.sh -b -p /ext3/miniconda3
fi

echo Updating conda
conda update -n base conda -y
[ ! -e "/ext3/miniconda3/$CONDA_ENV" ] && conda create -n $CONDA_ENV $CONDA_PYTHON  -y
conda install pip -y --upgrade

echo creating startup script
cat > /ext3/env << EOF2""" + r"""
#!/bin/bash
export PATH=/ext3/miniconda3/bin:\\\$PATH
source /ext3/miniconda3/etc/profile.d/conda.sh -y
conda activate $CONDA_ENV
echo "hello :) you're using:" "\\\$(which python)" "\\\$(python --version)"
EOF2

chmod +x /ext3/env
echo '/bin/bash --init-file /ext3/env' > /ext3/bash.env
chmod +x /ext3/bash.env
    """)


    header('installing packages in conda')

    sing(overlay, sif, f"""
REQUIREMENTS={requirements or ''}
INSTALLS={install or ''}
EDITABLE={package or ''}
""" + """
echo Doing installs...

echo requirements file: "$REQUIREMENTS"
if [ ! -z "$REQUIREMENTS" ]; then
    echo 'installing pip requirements file:' "$REQUIREMENTS"
    pip install -r $REQUIREMENTS
fi

echo packages to install: "$INSTALLS"
if [ ! -z "$INSTALLS" ]; then
    echo 'installing custom conda packages:' "$INSTALLS"
    conda install $INSTALLS
fi

echo project package: "$EDITABLE"
if [ ! -z "$EDITABLE" ]; then
    echo 'installing python package:' "$EDITABLE"
    pip install -e $EDITABLE
fi

conda clean --all --yes
    """)

    cmd_alias = 'sing'
    cmd = sing_cmd(overlay, sif)
    with open(cmd_alias, 'w') as f:
        f.write(cmd)
    slurmjobs.util.make_executable(cmd_alias)

    header(f"to run a bash shell in the container:{cmd}\n  (alias: ``{cmd_alias}``)")


def sing(overlay, sif, code):
    subprocess.run(f"""
singularity exec --overlay {overlay} {sif} /bin/bash << EOF
[[ -f /ext3/env ]] && . /ext3/env
{code}
EOF
    """)

def sing_cmd(overlay, sif):
    return f"singularity exec --overlay {overlay} {sif} /bin/bash --init-file /ext3/env"


def header(*text):
    print()
    print('*****')
    print()
    print(*text)
    print()
    print('*****')
    print()



from .__version__ import __version__
from . import args
from .grid import *
from .core import *
from . import util
from .receipt import *


# cute little alias
Sing = Singularity

get_cli = args.Argument.get